/*
	ALARM_SENSOR

	Gestione di un sistema di allarme con 
	sensore PIR

	Funzionamento:
 1) Se viene rilevato un singolo movimento
    si accende il led GIALLO per almeno 3 secondi
    (ovvero per 3 secondi dal momento in cui il
    movimento è terminato)
 2) Se vengono rilevati almeno TRE eventi di cui
    al punto (1) all'interno dell'intervallo di
    15 secondi il sistema entra in allarme ==>
    si accende quindi il led ROSSO (e non il GIALLO)
    e si accende il buzzer (suona) per almeno 5 secondi
    (ovvero 5 secondi da quando cessa il movimento)
 3) Se non ci sono ALLARMI né PREALLARMI rimane
    acceso il solo led BLUE
    NOTA:
    Durante lo sviluppo del programma ci siamo accorti
    che se il sensore rileva un "movimento", mentre 
    correttamente si attiva il segnale di preallarme e
    viene incrementato il conteggio dei preallarmi, e
    se tale "movimento" non termina mai ossia il "ladro"
    si muove continuamente si ha che il sistema non 
    entra MAI in stato di ALLARME.
    Pertanto in questa versione è stato corretto 
    questo problema...
*/

#define PIR_INPUT     2
#define LED_RED      10         
#define LED_YELLOW   11
#define LED_BLUE     12
#define BUZZER       13
// Tempi di attivazione ALLARME / PREALLARME (in msec)
#define TIME_PREALLARME	        3000
#define NUM_PREALLARMI             3
#define TIME_ALLARME            5000
#define TIME_LETTURA_PIR         100
#define TIME_RESET_PREALLARMI  15000
#define TIME_FORCE_ALLARME     10000

#define IN_ATTIVO 		0
#define IN_NON_ATTIVO 	1

#define OUT_ATTIVO 		1
#define OUT_NON_ATTIVO 	0

#define PIR_ATTIVO      1
#define PIR_NON_ATTIVO  0

bool newMovimento, oldMovimento;
bool Allarme, PreAllarme, ImpiantoOK;
int CountPreAllarmi;

unsigned long long TimeRefreshSensore;
unsigned long long TimeAllarme;
unsigned long long TimePreAllarme;
unsigned long long TimePrimoEvento;
unsigned long long TimeUltimoEvento;
long long deltaTimePrimoEvento;
long long deltaTimeUltimoEvento;

void setup() 
{
  // Definizione pin di IN / OUT
  pinMode(BUZZER, OUTPUT);
  pinMode(LED_RED, OUTPUT); 
  pinMode(LED_YELLOW, OUTPUT); 
  pinMode(LED_BLUE, OUTPUT); 
  pinMode(PIR_INPUT, INPUT); 
  // Eventuale seriale di DEBUG
  Serial.begin(9600);
  // Azzeramento variabili
  newMovimento = oldMovimento = PIR_NON_ATTIVO;
  Allarme = PreAllarme = ImpiantoOK = false;
  TimeRefreshSensore = TimeAllarme = TimePreAllarme = 0;
  TimePrimoEvento = TimeUltimoEvento = 0;
  deltaTimePrimoEvento = deltaTimeUltimoEvento = 0;
  CountPreAllarmi = 0;
  // Azzeramento uscite
  digitalWrite( BUZZER, OUT_NON_ATTIVO ); 
  digitalWrite( LED_RED, OUT_NON_ATTIVO ); 
  digitalWrite( LED_YELLOW, OUT_NON_ATTIVO ); 
  digitalWrite( PIR_INPUT, OUT_NON_ATTIVO ); 
}
 
void loop()
{
  //----------------------------------------------- 
  // Lettura INGRESSI
  //-----------------------------------------------     
  // Verifica se siamo nel tempo di lettura dell'ingresso PIR
  // (ogni 100 msec)
  if( millis() > TimeRefreshSensore )
  {  
    // Legge il sensore
  	newMovimento = digitalRead(PIR_INPUT);    
    // Se è cambiato lo stato (rilevato una variazione nello 
    // stato dell'uscita del sensore)
    if( newMovimento != oldMovimento )
    {
      if( newMovimento == PIR_ATTIVO )
      {
        // In questa variabile viene memorizzato
        // l'istante in cui INIZIA un evento di movimento
        TimePrimoEvento = millis();
        // Viene incrementato il conteggio dei preallarmi
        // ad ogni inizio di un nuovo movimento
        ++CountPreAllarmi;      
      }        
    }
    // Memorizza stato attuale per prossimo confronto
    oldMovimento = newMovimento;
    if( newMovimento == PIR_ATTIVO )
    {    
      // Viene utilizzato il valore attuale del tempo
      // per precalcolare il timeout del segnale di 
      // allarme e di preallarme (i tempi partono sempre
      // dall'ultimo movimento visto dal sensore ovvero
      // a partire dalla "fine" del movimento)
      TimePreAllarme = millis() + TIME_PREALLARME;        
      if( CountPreAllarmi >= NUM_PREALLARMI )
      {
        TimeAllarme = millis() + TIME_ALLARME;
      }  
    }
    // Se l'evento di movimento è iniziato e sta
    // continuando da oltre 10 secondi, ovvero il "ladro"
    // sta continuando a "lavorare" all'interno dell'area
    // controllata, in ogni caso viene attivato il
    // segnale di allarme senza utilizzare il conteggio
    // dei preallarmi all'interno dei 15 secondi
    if( newMovimento == PIR_ATTIVO )
    {
      // Calcola la differenza di tempo fra adesso
      // e l'inizio del movimento
      deltaTimePrimoEvento = millis() - TimePrimoEvento;
      // Non voglio visualizzare tempi negativi
      if( deltaTimePrimoEvento < 0 ) deltaTimePrimoEvento = 0;
      // Se il movimento sta durando da oltre 10 secondi
      // allora attiva l'allarme
      if( deltaTimePrimoEvento >= TIME_FORCE_ALLARME )
      {
        // Forza sicuramente la situazione di allarme
        CountPreAllarmi = NUM_PREALLARMI + 1;
      }
    }
    // Calcola tempo trascorso da ultimo evento di
    // movimento (la variabile viene continuamente 
    // aggiornata finché dura il movimento, rimane
    // quindi in memoria sempre l'istante dell'ultimo
    // movimento visto). Viene visualizzato a scopo di debug
    // ma, soprattutto, utilizzato per la temporizzazione
    // dei 15 secondi..
    if( newMovimento == PIR_ATTIVO ) TimeUltimoEvento = millis();
    // Calcola di volta in volta il tempo trascorso
    deltaTimeUltimoEvento = millis() - TimeUltimoEvento;
    // Non voglio visualizzare tempi negativi
    if( deltaTimeUltimoEvento < 0 ) deltaTimeUltimoEvento = 0;    
    // Se sono passati oltre 15 secondi dal precedente
    // preallarme allora azzera il contatore di preallarmi
    if( deltaTimeUltimoEvento >= TIME_RESET_PREALLARMI )
    {
      CountPreAllarmi = 0;
    }
    // Ricarica il timeout per la prossima lettura
    TimeRefreshSensore = millis() + TIME_LETTURA_PIR;   
    // Utilizza lo stesso timing per il debug su seriale
    Serial.print("MOVIM=");
    Serial.print(newMovimento);
    Serial.print(", tPRIMO=");
    Serial.print( (long) deltaTimePrimoEvento );      
    Serial.print(", tULTIMO=");
    Serial.print( (long) deltaTimeUltimoEvento );    
    Serial.print(", COUNT=");
    Serial.print( CountPreAllarmi); 
    Serial.println("");    
  }
  //----------------------------------------------- 
  // Elaborazione DATI
  //----------------------------------------------- 
  // Controllo su timeout durata segnale di preallarme
  if( millis() > TimePreAllarme )
  {
    PreAllarme = false;
  }
  else
  {
    PreAllarme = true;
  }
  // Controllo su timeout durata segnale di allarme
  if( millis() > TimeAllarme )
  {
    // Una volta attivato l'allarme azzera il 
    // conteggio dei preallarmi (solo alla
    // scadenza del timeout a seguito dell'
    // attivazione di un allarme)
    if( Allarme == true ) CountPreAllarmi = 0;
    Allarme = false;    
  }
  else
  {
    Allarme = true;
    // Il segnale di allarme ha priorità su quello
    // di preallarme per ci viene spento il segnale
    // a priorità inferiore
    PreAllarme = false;
  } 
  // Se non c'erano né Allarme, né PreAllarme
  // allora il Sistema è OK
  if( Allarme || PreAllarme )
    ImpiantoOK = false;
  else
    ImpiantoOK = true;
  //----------------------------------------------- 
  // Scrittura USCITE
  //-----------------------------------------------     
  // Accende gli opportuni LED / BUZZER
  if( Allarme ) 
    digitalWrite( LED_RED, OUT_ATTIVO );
  else
    digitalWrite( LED_RED, OUT_NON_ATTIVO );
  /*
  // NOTA: il buzzer è molto noioso ed è 
  // temporaneamente disattivato
  if( Allarme ) 
    digitalWrite( BUZZER, OUT_ATTIVO );
  else
    digitalWrite( BUZZER, OUT_NON_ATTIVO ); 
  */
  //
  if( PreAllarme ) 
      digitalWrite( LED_YELLOW, OUT_ATTIVO );
    else
      digitalWrite( LED_YELLOW, OUT_NON_ATTIVO );
  //
  if( ImpiantoOK ) 
    digitalWrite( LED_BLUE, OUT_ATTIVO );
  else
    digitalWrite( LED_BLUE, OUT_NON_ATTIVO );    
  
}





